

/*  File: adpcm.c
    Description: Routine to convert linear samples to 4-bit Oki ADPCM2
    format.
    Original Description: Routines to convert 12 bit linear samples to the
       Dialogic or Oki ADPCM coding format.
       I copied the algorithms out of the book "PC Telephony - The
       complete guide to designing, building and programming systems
       using Dialogic and Related Hardware" by Bob Edgar. pg 272-276.

*/

# include "adpcm.h"


#define MAXSAMP 64 //I'm not sure about this, but it seems to work OK...

static short step_size[MAXSAMP];
/*
* one function local to this file only.
*/
short step_adjust ( char );

/*
* Initialze the data used by the coder. No precalced table is used cause
* this should take no time at all on a modern PC
*/
void adpcm_init( struct adpcm_status *stat ) {
    int i;
    float j,oj;
    stat->last = 0;
    stat->step_index = 0;
    //Calculate step table
    oj=64;
    for (i=0; i<MAXSAMP; i++) {
	j=(float)oj*(float)1.08;
	oj=j;
	step_size[i]=j;
    }
    return;
}

/*
* Encode linear to ADPCM
*/
char adpcm_encode( short samp, struct adpcm_status *stat ) {
    short code;
    short diff, E, SS;

    /* printf( "%d\t", samp );
    */
    SS = step_size[stat->step_index];
    code = 0x00;
    if( (diff = samp - stat->last) < 0 )
        code = 0x08;
    E = diff < 0 ? -diff : diff;
    if( E >= SS ) {
        code = code | 0x04;
        E -= SS;
    }
    if( E >= SS/2 ) {
        code = code | 0x02;
        E -= SS/2;
    }
    if( E >= SS/4 ) {
        code = code | 0x01;
    }
/*    stat->step_index += step_adjust( code );
    if( stat->step_index < 0 ) stat->step_index = 0;
    if( stat->step_index > 48 ) stat->step_index = 48;
*/
    /*
    * Use the decoder to set the estimate of last sample.
    * It also will adjust the step_index for us.
    */
    stat->last = adpcm_decode(code, stat);
    return( code );
}

/*
* Decode Linear to ADPCM
*/
short adpcm_decode( char code, struct adpcm_status *stat ) {
    short diff, E, SS, samp;

    /* printf( "%x\t", code );
    */
    SS = step_size[stat->step_index];
    E = SS/8;
    if ( code & 0x01 )
        E += SS/4;
    if ( code & 0x02 )
        E += SS/2;
    if ( code & 0x04 )
        E += SS;
    diff = (code & 0x08) ? -E : E;
    samp = stat->last + diff;

    /*
    *  Clip the values to +(2^11)-1 to -2^11. (12 bits 2's
    *  compelement)
    *  Note: previous version errantly clipped at +2048, which could
    *  cause a 2's complement overflow and was likely the source of
    *  clipping problems in the previous version.  Thanks to Frank
    *  van Dijk for the correction.  TLB 3/30/04
    */
    
    if( samp > 16384 )
    {
      samp = 16384;
    }
    if( samp < -16384 )
    {
      samp = -16384;
    }

    stat->last = samp;
    stat->step_index += step_adjust( code );
    if( stat->step_index < 0 ) stat->step_index = 0;
    if( stat->step_index > MAXSAMP-1 ) stat->step_index = MAXSAMP-1;

    return( samp );
}

/*
* adjust the step for use on the next sample.
*/
short step_adjust ( char code ) {
    switch( code & 0x07 ) {
    //Originally, this was -1 -1 -1 -1 2 3 6 8 for adpcm2.
    //The following values seem to be spot-on for 4bit ADPCM2.
        case 0x00:
            return(-2);
            break;
        case 0x01:
            return(-2);
            break;
        case 0x02:
            return(-2);
            break;
        case 0x03:
            return(-2);
            break; //2 3 6 8
        case 0x04:
            return(2);
            break;
        case 0x05:
            return(6);
            break;
        case 0x06:
            return(9);
            break;
        case 0x07:
            return(11);
            break;
    }
}
